import argparse

from SurikovDB.DBMS import DBMS

parser = argparse.ArgumentParser(description="SurikovDB DBMS", prog='SurikovDB')
subparsers = parser.add_subparsers()


def list_db(args):
    db_list = DBMS.get_data_base_list()
    for i in db_list:
        print(f"Name: {i['name']}, path: {i['path']}")


def create_db(args):
    DBMS.create_data_base(args.path, args.database_name)


def drop_db(args):
    DBMS.drop_data_base(args.database_name)


def start_db(args):
    db = DBMS.get_data_base(args.database_name)
    host = args.host[0] if args.host else args.host
    port = args.port[0] if args.port else args.port
    db.run(host, port)


list_db_parser = subparsers.add_parser('list_db', help='List all databases')
list_db_parser.set_defaults(func=list_db)

create_db_parser = subparsers.add_parser('create_db', help='Create database')
create_db_parser.add_argument('database_name', help='Database name')
create_db_parser.add_argument('path', help='Path to folder to store database')
create_db_parser.set_defaults(func=create_db)

drop_db_parser = subparsers.add_parser('drop_db', help='Drop database')
drop_db_parser.add_argument('database_name', help='Database name')
drop_db_parser.set_defaults(func=drop_db)

start_db_parser = subparsers.add_parser('start_db', help='Start database http server')
start_db_parser.add_argument('database_name', help='Database name')
start_db_parser.add_argument('-H', dest='host', help='Host', nargs=1, default=None)
start_db_parser.add_argument('-P', dest='port', help='Port', nargs=1, default=None)
start_db_parser.set_defaults(func=start_db)


def main():
    args = parser.parse_args()
    args.func(args)




import struct
from struct import calcsize, unpack_from
from SurikovDB.DDL.Column import Column, ColumnFactory
from SurikovDB.constants import *


class TableMetaData:
    def __init__(self, name: str, column_list: list[Column]):
        if len(name.encode()) > calcsize(TABLE_NAME_F):
            raise Exception(f"Table name '{name}' more then table name max size({calcsize(TABLE_NAME_F)})")

        self._name = name
        self._column_list = column_list
        self._column_count = len(column_list)
        self._check_row_size()  # throw exception
        self._check_ddl_size()  # throw exception

    @property
    def row_struct_format(self) -> str:
        return f'{ROW_IS_DELETED_F}' + ''.join([c.struct_format for c in self._column_list])

    @property
    def name(self):
        return self._name

    @property
    def column_list(self) -> list[Column]:
        return self._column_list

    @property
    def encode_ddl(self) -> bytes:

        ddl = [
            self._name.encode(),
            len(self._column_list)
        ]

        for c in self._column_list:
            ddl.extend([c.name.encode(), c.code])

        return struct.pack(self._get_ddl_struct_format(), *ddl)

    @classmethod
    def from_encode_ddl(cls, encode_ddl: bytes) -> 'TableMetaData':
        ptr = 0

        table_name = unpack_from(TABLE_NAME_F, encode_ddl)[0].strip(b'\x00').decode()
        ptr = calcsize(TABLE_NAME_F)

        column_count = unpack_from(COLUMN_COUNT_F, encode_ddl, ptr)[0]
        ptr += calcsize(COLUMN_COUNT_F)

        column_list = []
        for i in range(column_count):
            col_name = unpack_from(COLUMN_NAME_F, encode_ddl, ptr)[0].strip(b'\x00').decode()
            ptr += calcsize(COLUMN_NAME_F)
            col_code = unpack_from(COLUMN_CODE_F, encode_ddl, ptr)[0]
            ptr += calcsize(COLUMN_CODE_F)
            column_list.append(ColumnFactory.column_from_code(col_name, col_code))

        return cls(table_name, column_list)

    def _check_row_size(self) -> None:
        f = self.row_struct_format
        size = calcsize(f)
        if size > BLOCK_SIZE - calcsize(ROW_COUNT_F):
            raise Exception('Row size more then block size.')

    def _check_ddl_size(self) -> None:
        if (calcsize(self._get_ddl_struct_format()) + calcsize(TABLE_DDL_SIZE_F)) > BLOCK_SIZE:
            raise Exception(f"Table ddl more then block size")

    def _get_ddl_struct_format(self) -> str:
        struct_format = f'{TABLE_NAME_F}{COLUMN_COUNT_F}' + (f'{COLUMN_NAME_F}{COLUMN_CODE_F}' * self._column_count)
        return struct_format

    @property
    def json(self):
        return {
            'table_name': self._name,
            'columns': [str(i) for i in self._column_list]
        }

    def __str__(self):
        column_list_str = '\n\t'.join(map(str, self._column_list))
        return f"{self._name}:\n\t{column_list_str}"



from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.Block import Block
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.JSONQLException import JSONQLException


class DropTable(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "table_name": {"type": "string"},
        },
        "additionalProperties": False,
        "required": ["type", "table_name"]
    }

    @classmethod
    def from_json(cls, json: Any) -> 'DropTable':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)

        table_name = json['table_name']
        return cls(table_name)

    def __init__(self, table_name: str):
        self._table_name = table_name
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        for table_meta_data, block_index in data_base_storage.table_meta_data_gen():
            if table_meta_data.name == self._table_name:
                b = data_base_storage.read_block(block_index)
                yield b
                b.free()
                data_base_storage.write_block(b)
                break
        else:
            raise Exception('Table is not exist')

    @property
    def result(self):
        return self._result



import os
from time import time
from typing import Union

from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.JSONQLParser import JSONQLParser
from SurikovDB.RowSet import RowSet
from SurikovDB.Server.server import DataBaseServer
from SurikovDB.Transaction import Transaction
from SurikovDB.constants import *


class DataBase:
    def __init__(self, path: str):
        self._data_base_storage = DataBaseStorage(path)
        self._name = os.path.split(path)[-1]
        self._path = path
        # self._recover_data()

    def query(self, json: Union[dict, list]) -> dict:
        start_time = time()
        result = {}
        try:
            command_list = JSONQLParser.parse(json)
            tr = Transaction(command_list, self._data_base_storage)
            try:
                tr.execute()
            except Exception as e:
                tr.rollback()
                raise e

            result = tr.result

            if isinstance(result, RowSet):
                result = result.json

            result = {
                'result': result,
            }

        except Exception as e:
            result = {
                'error': str(e)
            }

        finally:
            end_time = time()
            result['execution_time'] = end_time - start_time
            return result

    def get_table_list(self):
        table_list = []
        for i in self._data_base_storage.table_meta_data_gen():
            table_list.append(i.table_meta_data.json)
        return table_list

    def _recover_data(self):
        data_base_dir_path = os.path.join(*os.path.split(self._path)[0:-1])
        transaction_file_list = [i for i in os.listdir(data_base_dir_path) if i.startswith('tlog-')]

        transaction_file_list = sorted(transaction_file_list, key=lambda i: float(i[5:]), reverse=True)

        for transaction_file in transaction_file_list:
            transaction_file_path = os.path.join(data_base_dir_path, transaction_file)
            transaction = Transaction.from_transaction_file_path(self._data_base_storage, transaction_file_path)
            transaction.rollback()

    def __del__(self):
        self._data_base_storage.file.close()

    def run(self, host=None, port=None):
        dbs = DataBaseServer(self, host=host, port=port)
        dbs.run()

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    @property
    def data_base_storage(self):
        return self._data_base_storage



import json
import os
from typing import NoReturn
import shutil
from SurikovDB.constants import *
import platformdirs

from SurikovDB.DataBase import DataBase


class DBMS:
    data_base_list_file = os.path.join(platformdirs.user_data_dir('SurikovDB'), 'data_base_list.json')
    os.makedirs(os.path.dirname(data_base_list_file), exist_ok=True)

    if not os.path.exists(data_base_list_file):
        with open(data_base_list_file, 'x') as f:
            json.dump([], f)

    @staticmethod
    def get_data_base(name):
        data_base_list = DBMS.get_data_base_list()

        for i in data_base_list:
            if i['name'] == name:
                data_base = i
                break
        else:
            raise Exception('This database is not exists')
        return DataBase(os.path.join(data_base['path'], name))

    @staticmethod
    def create_data_base(path: str, name: str) -> DataBase:
        with open(DBMS.data_base_list_file, 'r') as f:
            data_base_list = json.load(f)

        db_path = os.path.join(path, name)

        if name in [os.path.split(path)[-1] for path in data_base_list]:
            raise Exception('Database name must be unique')

        os.mkdir(db_path)
        data_base_list.append(db_path)

        with open(DBMS.data_base_list_file, 'w') as f:
            json.dump(data_base_list, f)

        return DataBase(os.path.join(db_path, name))

    @staticmethod
    def drop_data_base(name: str) -> NoReturn:
        data_base_list = DBMS.get_data_base_list()

        for i in data_base_list:
            if i['name'] == name:
                data_base = i
                break
        else:
            raise Exception('This database is not exists')

        if os.path.exists(data_base['path']):
            shutil.rmtree(data_base['path'])

        data_base_list = [i['path'] for i in data_base_list if i['name'] != name]
        with open(DBMS.data_base_list_file, 'w') as f:
            json.dump(data_base_list, f)

    @staticmethod
    def get_data_base_list() -> list[dict]:
        with open(DBMS.data_base_list_file, 'r') as f:
            data_base_list = json.load(f)

        return [{'name': os.path.split(path)[-1], 'path': path} for path in data_base_list]




import abc
from abc import ABCMeta
from struct import calcsize
from typing import Any

from jsonschema import Draft202012Validator

from SurikovDB.JSONQLException import JSONQLException
from SurikovDB.constants import *


class Column(metaclass=ABCMeta):

    def __init__(self, name: str):
        if len(name.encode()) > calcsize(COLUMN_NAME_F):
            raise Exception(f"Column name '{name}' more then column name max size({calcsize(COLUMN_NAME_F)})")
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    @abc.abstractmethod
    def struct_format(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def code(self) -> int:
        pass

    @abc.abstractmethod
    def is_valid_value(self, value: Any) -> bool:
        pass


class ColumnChar(Column):
    json_schema = {
        "minItems": 3,
        "maxItems": 3,
        "type": "array",
        "prefixItems": [
            {"type": "string"},
            {"enum": ["char"]},
            {"type": "number"}
        ],
    }

    @classmethod
    def from_json(cls, json: Any) -> 'ColumnChar':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)
        return cls(json[0], json[2])

    def __init__(self, name: str, len_bytes: int):
        super().__init__(name)
        self._struct_format = 's'

        if len_bytes < 1:
            raise Exception('Incorrect char length')

        self._len_bytes = len_bytes

    def is_valid_value(self, value: Any) -> bool:
        if isinstance(value, str) and len(value.encode()) <= self._len_bytes:
            return True
        else:
            return False

    @property
    def struct_format(self) -> str:
        return f'{self._len_bytes}{self._struct_format}'

    @property
    def code(self) -> int:
        return self._len_bytes

    def __str__(self):
        return f'{self.name} - char({self._len_bytes})'


class ColumnLong(Column):
    json_schema = {
        "minItems": 2,
        "maxItems": 2,
        "type": "array",
        "prefixItems": [
            {"type": "string"},
            {"enum": ["long"]}
        ],

    }

    @classmethod
    def from_json(cls, json: Any) -> 'ColumnLong':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)
        return cls(json[0])

    def __init__(self, name: str):
        super().__init__(name)
        self._struct_format = 'l'

    def is_valid_value(self, value: Any) -> bool:
        if isinstance(value, int) and (-2147483648 <= value <= 2147483647):
            return True
        else:
            return False

    @property
    def struct_format(self) -> str:
        return self._struct_format

    @property
    def code(self) -> int:
        return -1

    def __str__(self):
        return f'{self.name} - long'


class ColumnFactory:

    @staticmethod
    def column_from_code(name: str, code: int) -> Column:
        if code > 0:
            return ColumnChar(name, code)
        elif code == -1:
            return ColumnLong(name)
        else:
            raise Exception('Incorrect code')

    @staticmethod
    def column_from_json(json: Any) -> Column:
        if json[1] == 'char':
            return ColumnChar.from_json(json)
        elif json[1] == 'long':
            return ColumnLong.from_json(json)
        else:
            raise NotImplementedError



import typing
from typing import Generator

from SurikovDB.BlockStorage import BlockStorage
from SurikovDB.RowSet import RowSet
from SurikovDB.DDL.TableMetaData import TableMetaData
from SurikovDB.Block import TableMetaDataBlock, DataBlock
from SurikovDB.constants import *


class DataBaseStorage(BlockStorage):
    TABLE_META_DATA_BLOCK_COUNT = 10
    table_meta_data_gen_result = typing.NamedTuple('table_meta_data_gen_result',
                                                   [('table_meta_data', TableMetaData), ('block_index', int)])

    def __init__(self, path: str):
        super().__init__(path, BLOCK_SIZE)

        if self.block_count == 0:
            for _ in range(self.TABLE_META_DATA_BLOCK_COUNT):
                self.allocate_block()

    def scan_table_data_block_gen(self,
                                  table_meta_data: TableMetaData,
                                  table_meta_data_block_index: int) -> Generator[DataBlock, None, None]:

        row_format = table_meta_data.row_struct_format
        data_block_table_meta_data = TableMetaDataBlock.from_block(self.read_block(table_meta_data_block_index))

        data_block_pointer_lvl1 = DataBlock.from_block(
            self.read_block(data_block_table_meta_data.data_block_pointer_lvl1), POINTER_F)

        for i in data_block_pointer_lvl1.read_rows():
            data_block_pointer_lvl2_list = DataBlock.from_block(self.read_block(i[0]), POINTER_F).read_rows()
            for j in data_block_pointer_lvl2_list:
                data_block_table_data = DataBlock.from_block(self.read_block(j[0]), row_format)
                yield data_block_table_data

    def scan(self, table_name: str) -> RowSet:
        for table_meta_data, block_index in self.table_meta_data_gen():
            if table_meta_data.name == table_name:
                break
        else:
            raise Exception('Table is not exist')

        def row_gen() -> Generator[ROW_TYPE, None, None]:
            for table_data_block in self.scan_table_data_block_gen(table_meta_data, block_index):
                row_list = table_data_block.read_rows()
                for k in row_list:
                    if not k[0]:
                        yield k[1:]

        return RowSet([c.name for c in table_meta_data.column_list], row_gen, [f'{table_name}'])

    def table_meta_data_gen(self) -> Generator[table_meta_data_gen_result, None, None]:
        for block_index in range(self.TABLE_META_DATA_BLOCK_COUNT):
            b = TableMetaDataBlock.from_block(self.read_block(block_index))

            if b.is_empty:
                continue

            yield self.table_meta_data_gen_result(b.get_table_meta_data(), block_index)

    def get_table_meta_data(self, table_name: str) -> table_meta_data_gen_result:
        for i in self.table_meta_data_gen():
            if i.table_meta_data.name == table_name:
                return i
        else:
            raise Exception('Table is not exist')



from typing import Generator, Callable

from SurikovDB.Expression import Expression
from SurikovDB.constants import *


class RowSet:
    def __init__(self, column_name_list: list[str], row_gen: Callable[[], Generator[ROW_TYPE, None, None]], info: list):
        self._column_name_list = column_name_list
        self._row_gen = row_gen
        self._info = info

    def where(self, filter_exp: Expression) -> 'RowSet':
        filter_exp_func, filter_exp_arg_name_list = filter_exp.parse()

        def row_gen() -> Generator[ROW_TYPE, None, None]:
            for row in self.row_gen:
                row_map = self._row_map(row)
                if filter_exp_func(**row_map):
                    yield row

        info = self._info + [f'where: {filter_exp.value}']
        return RowSet(self.column_name_list, row_gen, info)

    def select(self, exp_list: list[Expression]) -> 'RowSet':

        exp_func_list = [exp.parse()[0] for exp in exp_list]

        j = 0
        column_name_list = []
        for exp in exp_list:
            if exp.value in self._column_name_list:
                column_name_list.append(exp.value)
            else:
                column_name_list.append(f'col_{j}')
                j += 1

        def row_gen() -> Generator[ROW_TYPE, None, None]:
            for row in self.row_gen:
                row_map = self._row_map(row)
                row_result = []
                for exp_func in exp_func_list:
                    row_result.append(exp_func(**row_map))
                yield tuple(row_result)

        info = self._info + [f'select: {[exp.value for exp in exp_list]}']
        return RowSet(column_name_list, row_gen, info)

    def join(self, row_set: 'RowSet', on_exp: Expression) -> 'RowSet':

        on_exp_func, on_exp_arg_name_list = on_exp.parse()
        row_set_source_table = row_set.info[0]
        source_table = self.info[0]
        column_name_list = [f'{source_table}.{c}' if '.' not in c else c for c in self._column_name_list] + \
                           [f'{row_set_source_table}.{c}' if '.' not in c else c for c in row_set.column_name_list]

        def row_gen() -> Generator[ROW_TYPE, None, None]:
            for row1 in self.row_gen:
                for row2 in row_set.row_gen:
                    row = row1 + row2
                    row_map = self._row_map(row, column_name_list)
                    if on_exp_func(**row_map):
                        yield row

        info = self._info + [f'join: {row_set.info} on {on_exp.value}']
        return RowSet(column_name_list, row_gen, info)

    def _row_map(self, row: ROW_TYPE, column_name_list: list[str] = None) -> dict[str, DB_TYPE]:
        if column_name_list is None:
            column_name_list = self._column_name_list
        return {c: v for c, v in zip(column_name_list, row)}

    @property
    def column_name_list(self) -> list[str]:
        return self._column_name_list

    @property
    def row_gen(self) -> Generator[ROW_TYPE, None, None]:
        return self._row_gen()

    @property
    def info(self):
        return self._info

    @property
    def json(self):
        return {
            'columns': self.column_name_list,
            'data': [row for row in self.row_gen]
        }



from typing import Optional

from aiohttp import web



class DataBaseServer:

    def __init__(self, data_base, host: Optional[str] = None, port: Optional[int] = None):
        self._data_base = data_base
        if host is None:
            host = 'localhost'

        if port is None:
            port = '3471'

        self._port = port
        self._host = host

    def run(self):
        app = web.Application()
        app.add_routes([web.get('/', self.get_data_base_info_handler),
                        web.post('/', self.post_data_base_query_handler)])
        web.run_app(app, host=self._host, port=self._port)

    async def get_data_base_info_handler(self, request):
        response = {
            'data_base_name': self._data_base.name,
            'tables': self._data_base.get_table_list()
        }

        return web.json_response(response)

    async def post_data_base_query_handler(self, request):
        json = await request.json()
        response = self._data_base.query(json)
        return web.json_response(response)



from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.Block import Block
from SurikovDB.DML.InsertRow import InsertRow
from SurikovDB.DML.Select import Select
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.JSONQLException import JSONQLException
from SurikovDB.RowSet import RowSet


class Insert(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "table_name": {"type": "string"},
            "select": Select.json_schema
        },
        "additionalProperties": False,
        "required": ["type", "table_name", "select"]
    }

    @classmethod
    def from_json(cls, json: Any) -> 'Insert':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)

        table_name = json['table_name']
        select = Select.from_json(json['select'])
        return cls(table_name, select)

    def __init__(self, table_name: str, select: Select):
        self._table_name = table_name
        self._select = select
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        select = self._select
        for modified_block in select.execute(data_base_storage):
            yield modified_block

        row_set: RowSet = select.result

        row_count = 0
        for row in row_set.row_gen:
            insert_row = InsertRow(self._table_name, row)
            for modified_block in insert_row.execute(data_base_storage):
                yield modified_block
            row_count += 1

        self._result = row_count

    @property
    def result(self):
        return self._result



from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.Block import Block
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.Expression import Expression
from SurikovDB.JSONQLException import JSONQLException


class Select(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "select": {
                "type": "array",
                "items": Expression.json_schema},
            "from": {"type": "string"},
            "join": {
                "type": "object",
                "additionalProperties": Expression.json_schema
            },
            "where":
                Expression.json_schema

        },
        "additionalProperties": False,
        "required": ["type", "select", "from"]

    }

    @classmethod
    def from_json(cls, json: Any) -> 'Select':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)
        return cls(json)

    def __init__(self, json: any):
        self._json = json
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        json = self._json
        table_name = json['from']
        row_set_result = data_base_storage.scan(table_name)

        if 'join' in json:
            for t, exp in json['join'].items():
                row_set = data_base_storage.scan(t)
                row_set_result = row_set_result.join(row_set, Expression(exp))

        if 'where' in json:
            row_set_result = row_set_result.where(Expression(json['where']))

        if json['select'] != ['*']:
            row_set_result = row_set_result.select([Expression(e) for e in json['select']])

        self._result = row_set_result
        yield from []  # execute() need to be generator

    @property
    def result(self):
        return self._result



from jsonschema import ValidationError
from SurikovDB.constants import *

class JSONQLException(Exception):
    def __init__(self, error_list: list[ValidationError]):
        self.error_list = error_list
        self.message = 'JSONQL parse exception'
        super().__init__(self.message)

    def __str__(self):
        error_list_str = '\n'.join([error.message for error in self.error_list])
        return f"{self.message}:\n {error_list_str}"



from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.DDL.Column import ColumnLong, ColumnChar, ColumnFactory
from SurikovDB.JSONQLException import JSONQLException
from SurikovDB.DDL.TableMetaData import TableMetaData
from SurikovDB.Block import Block, TableMetaDataBlock, DataBlock
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.constants import POINTER_F


class CreateTable(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "table_name": {"type": "string"},
            "columns": {
                "type": "array",
                "items":
                    {"anyOf": [
                        ColumnLong.json_schema,
                        ColumnChar.json_schema,
                    ]},
                "minItems": 1,

            }
        },
        "additionalProperties": False,
        "required": ["type", "table_name", "columns"]

    }

    @classmethod
    def from_json(cls, json: Any) -> 'CreateTable':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)

        table_name = json['table_name']
        column_list = [ColumnFactory.column_from_json(c) for c in json['columns']]
        table_meta_data = TableMetaData(table_name, column_list)
        return cls(table_meta_data)

    def __init__(self, table_meta_data: TableMetaData):
        self._table_meta_data = table_meta_data
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        for t in data_base_storage.table_meta_data_gen():
            if self._table_meta_data.name == t.table_meta_data.name:
                raise Exception('Table name must be unique')


        for i in range(data_base_storage.TABLE_META_DATA_BLOCK_COUNT):
            b = data_base_storage.read_block(i)
            if b.is_empty:
                b = TableMetaDataBlock.from_block(b)
                break
        else:
            raise Exception('No space for table')

        data_block_pointer_lvl1 = DataBlock.from_block(data_base_storage.allocate_block(), POINTER_F)
        data_block_pointer_lvl2 = DataBlock.from_block(data_base_storage.allocate_block(), POINTER_F)
        data_block_table_data = data_base_storage.allocate_block()

        yield data_block_pointer_lvl2
        data_block_pointer_lvl2.write_row((data_block_table_data.idx,))

        yield data_block_pointer_lvl1
        data_block_pointer_lvl1.write_row((data_block_pointer_lvl2.idx,))

        yield b
        b.create_table(self._table_meta_data, data_block_pointer_lvl1.idx)

        data_base_storage.write_block(data_block_pointer_lvl1)
        data_base_storage.write_block(data_block_pointer_lvl2)
        data_base_storage.write_block(data_block_table_data)
        data_base_storage.write_block(b)

    @property
    def result(self):
        return self._result



from typing import Union, Tuple

BLOCK_SIZE = 4 * 1024

POINTER_F = 'L'
POINTER_COUNT_F = 'H'
ROW_COUNT_F = 'H'

TABLE_NAME_F = '50s'
COLUMN_COUNT_F = 'H'
COLUMN_NAME_F = '50s'
COLUMN_CODE_F = 'h'
TABLE_DDL_SIZE_F = 'H'
ROW_IS_DELETED_F = '?'


DB_TYPE = Union[bool, int, str]
ROW_TYPE = Tuple[DB_TYPE, ...]


from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.Block import Block
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.Expression import Expression
from SurikovDB.JSONQLException import JSONQLException


class Update(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "table_name": {"type": "string"},
            "set": {
                "type": "object",
                "additionalProperties": Expression.json_schema
            },
            "where": Expression.json_schema
        },
        "additionalProperties": False,
        "required": ["type", "table_name", "set"]
    }

    @classmethod
    def from_json(cls, json: Any) -> 'Update':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)

        table_name = json['table_name']
        if 'where' in json:
            filter_exp = Expression(json['where'])
        else:
            filter_exp = Expression(True)

        update_map = {}
        for c, exp in json['set'].items():
            update_map[c] = Expression(exp)

        return cls(table_name, filter_exp, update_map)

    def __init__(self, table_name: str, filter_exp: Expression, update_map: dict[str, Expression]):
        self._table_name = table_name
        self._filter_exp = filter_exp
        self._update_map = update_map
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        table_meta_data, block_index = data_base_storage.get_table_meta_data(self._table_name)

        column_map = {
            c.name: i + 1 for i, c in enumerate(table_meta_data.column_list)
        }
        column_map['.ROW_IS_DELETED'] = 0

        result = 0

        for data_block in data_base_storage.scan_table_data_block_gen(table_meta_data, block_index):
            row_list = data_block.read_rows()
            filter_exp_func, filter_exp_arg_name_list = self._filter_exp.parse()

            if not filter_exp_arg_name_list.issubset(column_map.keys()):
                raise Exception(
                    f'columns: {filter_exp_arg_name_list - set(column_map.keys())} is not exist in {self._table_name}')

            row_list_filtered = []
            for row_id, row in enumerate(row_list):
                filter_exp_arg = {i: row[column_map[i]] for i in filter_exp_arg_name_list}
                if not filter_exp_func(**filter_exp_arg):
                    continue

                row_list_filtered.append((row_id, row))

            row_list_filtered = [i for i in row_list_filtered if i[1][0] == False]

            if len(row_list_filtered) == 0:
                continue

            row_list_result = []
            for row_id, row in row_list_filtered:
                for column_name, update_exp in self._update_map.items():
                    update_exp_func, update_exp_arg_name_list = update_exp.parse()

                    if not update_exp_arg_name_list.issubset(column_map.keys()):
                        raise Exception(
                            f'columns: {filter_exp_arg_name_list - set(column_map.keys())} is not exist in {self._table_name}')

                    update_exp_arg = {i: row[column_map[i]] for i in update_exp_arg_name_list}
                    row = list(row)
                    row[column_map[column_name]] = update_exp_func(**update_exp_arg)

                for v, c in zip(row[1:], table_meta_data.column_list):
                    if not c.is_valid_value(v):
                        raise Exception(f'Column {c.name} - value {v} is not valid')

                row_list_result.append((row_id, row))

            yield data_block
            for row_id, row in row_list_result:
                result += 1
                data_block.write_row(row, row_id)

            data_base_storage.write_block(data_block)
        self._result = result

    @property
    def result(self):
        return self._result



from typing import Generator

from SurikovDB.Block import TableMetaDataBlock, DataBlock, Block
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.constants import ROW_TYPE, POINTER_F


class InsertRow(DataBaseCommand):
    def __init__(self, table_name: str, row: ROW_TYPE):
        self._table_name = table_name
        self._row = row
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        table_meta_data, block_index = data_base_storage.get_table_meta_data(self._table_name)

        if len(self._row) != len(table_meta_data.column_list):
            raise Exception(f'Invalid row size - row :{self._row}, column list - {table_meta_data.column_list}')

        for v, c in zip(self._row, table_meta_data.column_list):
            if not c.is_valid_value(v):
                raise Exception(f'Column {c.name} - value {v} is not valid')

        row = (False,) + self._row
        row_format = table_meta_data.row_struct_format

        data_block_table_meta_data = TableMetaDataBlock.from_block(data_base_storage.read_block(block_index))

        data_block_pointer_lvl1 = DataBlock.from_block(
            data_base_storage.read_block(data_block_table_meta_data.data_block_pointer_lvl1), POINTER_F)

        data_block_pointer_lvl2 = DataBlock.from_block(
            data_base_storage.read_block(data_block_pointer_lvl1.read_last_row()[0]),
            POINTER_F)

        data_block_table_data = DataBlock.from_block(
            data_base_storage.read_block(data_block_pointer_lvl2.read_last_row()[0]),
            row_format)

        if data_block_table_data.is_full:
            if data_block_pointer_lvl2.is_full:
                data_block_pointer_lvl2 = DataBlock.from_block(data_base_storage.allocate_block(), POINTER_F)
                yield data_block_pointer_lvl1
                data_block_pointer_lvl1.write_row((data_block_pointer_lvl2.idx,))
                data_base_storage.write_block(data_block_pointer_lvl1)

            data_block_table_data = DataBlock.from_block(data_base_storage.allocate_block(), row_format)
            yield data_block_pointer_lvl2
            data_block_pointer_lvl2.write_row((data_block_table_data.idx,))
            data_base_storage.write_block(data_block_pointer_lvl2)

        yield data_block_table_data
        data_block_table_data.write_row(row)
        data_base_storage.write_block(data_block_table_data)

    @property
    def result(self):
        return self._result



from setuptools import setup, find_packages

setup(
    name='SurikovDB',
    version='1.0.0',
    description='Simple relational database. Use JSONQL',
    author='Igor Surikov',
    author_email='i.surikov@outlook.com',
    url='https://github.com/IgorSurikov/SurikovDB',
    package_dir={'': 'src'},
    packages=find_packages('src', include=['SurikovDB', 'SurikovDB.*']),
    install_requires=[
        'jsonschema>=4.4.0',
        'aiohttp>=3.8.1',
        'platformdirs>=2.5.1'
    ],
    entry_points={
        'console_scripts': ['SurikovDB=SurikovDB.CLISurikovDB:main']
    },
)



import abc
from abc import ABCMeta
from typing import Generator, Any
from SurikovDB.constants import *

from SurikovDB.Block import Block
from SurikovDB.DataBaseStorage import DataBaseStorage


class DataBaseCommand(metaclass=ABCMeta):

    @abc.abstractmethod
    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        pass

    @property
    @abc.abstractmethod
    def result(self) -> Any:
        pass



from struct import unpack_from, pack, calcsize, iter_unpack
from typing import NoReturn

from SurikovDB.DDL.TableMetaData import TableMetaData
from SurikovDB.constants import *


class Block(bytearray):

    def __init__(self, data: bytes, index: int = 0, block_size: int = 0):
        self.block_size = block_size
        assert len(data) == self.block_size, 'Incorrect block size: {}'.format(len(data))
        super().__init__(data)

        self.idx = index

    @classmethod
    def empty(cls, index: int = 0, block_size: int = 0) -> 'Block':
        return cls(b'\0' * block_size, index, block_size)

    @property
    def is_empty(self) -> bool:
        return self == b'\0' * self.block_size

    def override(self, offset: int, data: bytes) -> None:
        assert offset + len(data) <= self.block_size

        self[offset:offset + len(data)] = data

    def free(self):
        self.override(0, b'\0' * self.block_size)


class TableMetaDataBlock(Block):

    def __init__(self, data: bytes, index: int = 0, block_size: int = 0):
        super().__init__(data, index, block_size)
        self._set_data_pointers()

    @classmethod
    def from_block(cls, block: Block) -> 'TableMetaDataBlock':
        return cls(bytes(block), block.idx, len(block))

    def _set_data_pointers(self) -> NoReturn:
        self._ddl_size = unpack_from(TABLE_DDL_SIZE_F, self, 0)[0]
        self._ddl_start_ptr = 0 + calcsize(TABLE_DDL_SIZE_F)
        self._data_block_pointer_lvl1_start_ptr = self._ddl_start_ptr + self._ddl_size

    def create_table(self, table_meta_data: TableMetaData, data_block_pointer_lvl1: int) -> NoReturn:
        ddl = table_meta_data.encode_ddl
        ddl = pack(f'{TABLE_DDL_SIZE_F}', len(ddl)) + ddl
        data_block_pointer_lvl1_encode = pack(POINTER_F, data_block_pointer_lvl1)

        if len(ddl) > self.block_size:
            raise Exception('DDL size is more than BLOCK_SIZE')

        self.override(0, ddl)
        self.override(len(ddl), data_block_pointer_lvl1_encode)
        self._set_data_pointers()

    def drop_table(self) -> NoReturn:
        self.free()

    @property
    def data_block_pointer_lvl1(self) -> int:
        return unpack_from(POINTER_F, self, self._data_block_pointer_lvl1_start_ptr)[0]

    def get_table_meta_data(self) -> TableMetaData:
        if self.is_empty:
            raise Exception('No Table Metadata')

        return TableMetaData.from_encode_ddl(self[self._ddl_start_ptr: self._ddl_start_ptr + self._ddl_size])


class DataBlock(Block):

    def __init__(self, data: bytes, index: int, row_format: str, block_size: int = 0):
        super().__init__(data, index, block_size)

        self._row_format = row_format
        self._set_data_pointers()

    @property
    def row_count_enable_to_insert(self) -> int:
        return (self.block_size - calcsize(ROW_COUNT_F) - (self._row_count * calcsize(self._row_format))) // calcsize(
            self._row_format)

    @property
    def is_full(self) -> bool:
        return self.row_count_enable_to_insert == 0

    def _set_data_pointers(self) -> NoReturn:
        self._row_count = unpack_from(f'{ROW_COUNT_F}', self, 0)[0]

        self._row_list_start_ptr = 0 + calcsize(f'{ROW_COUNT_F}')
        self._row_list_end_ptr = self._row_list_start_ptr + (self._row_count * calcsize(f'{self._row_format}'))

    @classmethod
    def from_block(cls, block: Block, row_format: str) -> 'DataBlock':
        return cls(bytes(block), block.idx, row_format, len(block))

    def write_row(self, row: ROW_TYPE, row_id: int = None) -> NoReturn:
        row = tuple([i.encode() if isinstance(i, str) else i for i in row])
        row_encode = pack(f'{self._row_format}', *row)

        if row_id is None:
            if self._row_list_end_ptr + len(row) > self.block_size:
                raise Exception('No space for row list')

            row_count_encode = pack(f'{ROW_COUNT_F}', self._row_count + 1)
            self.override(self._row_list_end_ptr, row_encode)
            self.override(0, row_count_encode)
            self._set_data_pointers()
        else:
            self.override(self._row_list_start_ptr + (row_id * calcsize(self._row_format)), row_encode)

    def read_rows(self) -> list[ROW_TYPE]:
        res = []
        for i in iter_unpack(self._row_format, self[self._row_list_start_ptr: self._row_list_end_ptr]):
            res.append(i)
        return [tuple([j.strip(b'\x00').decode() if isinstance(j, bytes) else j for j in i]) for i in res]

    def read_last_row(self) -> tuple:
        return self.read_rows()[::-1][0]



from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.Block import Block
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DML.Update import Update
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.Expression import Expression
from SurikovDB.JSONQLException import JSONQLException


class Delete(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "table_name": {"type": "string"},
            "where": Expression.json_schema
        },
        "additionalProperties": False,
        "required": ["type", "table_name"]
    }

    @classmethod
    def from_json(cls, json: Any) -> 'Delete':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)

        table_name = json['table_name']
        if 'where' in json:
            filter_exp = Expression(json['where'])
        else:
            filter_exp = Expression(True)
        return cls(table_name, filter_exp)

    def __init__(self, table_name: str, filter_exp: Expression):
        self._table_name = table_name
        self._filter_exp = filter_exp
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        update_map = {'.ROW_IS_DELETED': Expression(True)}

        update = Update(self._table_name, self._filter_exp, update_map)

        for modified_block in update.execute(data_base_storage):
            yield modified_block

        self._result = update.result

    @property
    def result(self):
        return self._result



from typing import Generator, Any

from jsonschema import Draft202012Validator

from SurikovDB.Block import Block
from SurikovDB.DML.InsertRow import InsertRow
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DataBaseStorage import DataBaseStorage
from SurikovDB.JSONQLException import JSONQLException
from SurikovDB.constants import ROW_TYPE


class InsertRows(DataBaseCommand):
    json_schema = {
        "type": "object",
        "properties": {
            "type": {"type": "string"},
            "table_name": {"type": "string"},
            "rows": {
                "type": "array",
                "items": {
                    "type": "array",
                    "items": {
                        "anyOf": [
                            {"type": "string"},
                            {"type": "number"},
                        ]}
                },
                "minItems": 1,
            }

        },
        "additionalProperties": False,
        "required": ["type", "table_name", "rows"]
    }

    @classmethod
    def from_json(cls, json: Any) -> 'InsertRows':
        errors = list(Draft202012Validator(cls.json_schema).iter_errors(json))
        if errors:
            raise JSONQLException(errors)

        table_name = json['table_name']
        row_list = [tuple(i) for i in json['rows']]
        return cls(table_name, row_list)

    def __init__(self, table_name: str, row_list: list[ROW_TYPE]):
        self._table_name = table_name
        self._row_list = row_list
        self._result = None

    def execute(self, data_base_storage: DataBaseStorage) -> Generator[Block, None, None]:
        for row in self._row_list:
            insert_row = InsertRow(self._table_name, row)
            for modified_block in insert_row.execute(data_base_storage):
                yield modified_block

        self._result = len(self._row_list)

    @property
    def result(self):
        return self._result



import datetime
import os
from struct import unpack_from, pack, calcsize
from typing import NoReturn, Generator, Optional

from SurikovDB import DataBaseStorage
from SurikovDB.Block import Block
from SurikovDB.BlockStorage import BlockStorage
from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.constants import *


class Transaction(BlockStorage):

    def __init__(self, command_list: list[DataBaseCommand],
                 data_base_storage: DataBaseStorage,
                 transaction_file_path: Optional[str] = None):

        if transaction_file_path is None:
            data_base_dir = os.path.split(data_base_storage.path)[0:-1]
            transaction_file_name = 'tlog-' + str(datetime.datetime.now().timestamp())
            transaction_file_path = os.path.join(*data_base_dir, transaction_file_name)

        super().__init__(transaction_file_path, BLOCK_SIZE + calcsize(POINTER_F))

        self._command_list = command_list
        self._result = None
        self._block_idx_list = [i.idx for i in self._block_gen()]
        self._path = transaction_file_path
        self._data_base_storage = data_base_storage

    @classmethod
    def from_transaction_file_path(cls, data_base_storage: DataBaseStorage, transaction_file_path: str):
        return cls([], data_base_storage, transaction_file_path)

    def __del__(self):
        self.file.close()
        os.remove(self.path)

    @property
    def result(self):
        return self._result

    def _save_database_storage_block(self, block: Block) -> NoReturn:
        if block.idx in self._block_idx_list:
            return

        block_target = self.allocate_block()
        block_index_encode = pack(f'{POINTER_F}', block.idx)
        block_target.override(0, block_index_encode)
        block_target.override(calcsize(f'{POINTER_F}'), bytes(block))
        self.write_block(block_target)
        self._block_idx_list.append(block.idx)

    def _block_gen(self) -> Generator[Block, None, None]:
        for i in range(self.block_count):
            block = self.read_block(i)
            database_storage_block_data = block[calcsize(f'{POINTER_F}'):]
            database_storage_block_idx = unpack_from(f'{POINTER_F}', block)[0]
            yield Block(database_storage_block_data, database_storage_block_idx, BLOCK_SIZE)

    def rollback(self) -> NoReturn:
        for i in self._block_gen():
            self._data_base_storage.write_block(i)

    def execute(self) -> NoReturn:

        for c in self._command_list:
            for modified_block in c.execute(self._data_base_storage):
                self._save_database_storage_block(modified_block)

        self._result = self._command_list[::-1][0].result



from typing import Union, Callable, Any, Set, Tuple
from SurikovDB.constants import *

class Expression:
    json_schema = {
        "anyOf": [
            {"type": "string"},
            {"type": "number"},
            {"type": "boolean"},
            {"type": "array", "minItems": 2},
        ]
    }

    math_function_map = {
        'add': lambda *x: x[0] + x[1],
        'sub': lambda *x: x[0] - x[1],
        'mul': lambda *x: x[0] * x[1],
        'div': lambda *x: x[0] / x[1],
        'eq': lambda *x: x[0] == x[1],
        'gt': lambda *x: x[0] > x[1],
        'ge': lambda *x: x[0] >= x[1],
        'lt': lambda *x: x[0] < x[1],
        'le': lambda *x: x[0] <= x[1],
        'ne': lambda *x: x[0] != x[1],
        'and': lambda *x: x[0] and x[1],
        'or': lambda *x: x[0] or x[1],
        'to_str': lambda *x: str(x[0]),
        'to_int': lambda *x: int(x[0]),
    }

    def __init__(self, value: Union[dict, int, str, list], context_args=None):
        if context_args is None:
            context_args = set()

        self.value = value
        self.context_args = context_args

    def parse(self) -> Tuple[Callable[..., Any], Set[str]]:
        v = self.value

        if isinstance(v, str):
            if v.startswith("'") and v.endswith("'"):
                v = v.strip("'")
                return lambda **x: v, self.context_args

            elif v in self.math_function_map:
                return lambda: {
                    'fn': self.math_function_map[f'{v}']
                }, self.context_args

            elif isinstance(v, bool):
                return lambda **x: v, self.context_args

            elif isinstance(v, str) and not v.startswith("'") and not v.endswith("'"):
                self.context_args.add(v)

                def f(**x):
                    try:
                        return x[f'{v}']
                    except KeyError as e:
                        raise Exception(f'No column: {e.args[0]}')

                return lambda **x: f(**x), self.context_args

            else:
                raise Exception(f'invalid string: {v}')

        if isinstance(v, int):
            return lambda **x: v, self.context_args

        if isinstance(v, list):
            try:
                func = Expression(v[0], self.context_args).parse()[0]()['fn']
            except (TypeError, KeyError):
                raise Exception(f'invalid function: {v}')

            args = [Expression(i, self.context_args).parse()[0] for i in v[1:]]
            return lambda **x: func(*[a(**x) for a in args]), self.context_args



from typing import Union

from jsonschema import Draft202012Validator

from SurikovDB.DataBaseCommand import DataBaseCommand
from SurikovDB.DDL.CreateTable import CreateTable
from SurikovDB.DDL.DropTable import DropTable
from SurikovDB.DML.Delete import Delete
from SurikovDB.DML.Insert import Insert
from SurikovDB.DML.InsertRows import InsertRows
from SurikovDB.DML.Select import Select
from SurikovDB.DML.Update import Update
from SurikovDB.JSONQLException import JSONQLException
from SurikovDB.constants import *

class JSONQLParser:
    json_schema = {
        "type": "object",
        "properties": {
            "type": {
                "enum": ["create_table", "drop_table", "select", "insert_rows", "insert", "delete", "update"]
            }
        },
        "required": ["type"]

    }

    @staticmethod
    def parse(json: Union[dict, list]) -> list[DataBaseCommand]:
        command_list = []
        command_list_json = []
        if isinstance(json, list):
            command_list_json = json
        else:
            command_list_json.append(json)

        for c_json in command_list_json:
            errors = list(Draft202012Validator(JSONQLParser.json_schema).iter_errors(c_json))
            if errors:
                raise JSONQLException(errors)

            command_type = c_json['type']

            command_type_map = {
                'create_table': CreateTable,
                'drop_table': DropTable,
                'select': Select,
                'insert_rows': InsertRows,
                'insert': Insert,
                'delete': Delete,
                'update': Update,
            }

            command = command_type_map[command_type].from_json(c_json)
            command_list.append(command)

        return command_list



import os
from typing import NoReturn, BinaryIO

from SurikovDB.Block import Block
from SurikovDB.constants import *


class BlockStorage:
    def __init__(self, path: str, block_size: int):
        self._block_size = block_size
        if not os.path.exists(path):
            f = open(path, 'x')
            f.close()

        self._path = path
        self._file = open(path, 'r+b', buffering=16 * self._block_size)

        file_size = os.path.getsize(path)

        if file_size % block_size:
            raise Exception('Incorrect BlockStorage file')
        else:
            self._block_count = file_size // self._block_size

    def read_block(self, index: int) -> Block:
        if index > self._block_count:
            raise Exception('Block is not exist')

        self._file.seek(index * self._block_size)
        block = Block(self._file.read(self._block_size), index, self._block_size)
        return block

    def write_block(self, block: Block) -> NoReturn:
        if len(block) != self._block_size:
            raise Exception('Block size is incorrect')

        if block.idx > self._block_count:
            raise Exception('Block is not exist')

        self._file.seek(block.idx * self._block_size)
        self._file.write(block)

    def allocate_block(self) -> Block:
        block = Block.empty(self._block_count, self._block_size)
        self._block_count += 1
        self.write_block(block)
        return block

    @property
    def block_count(self) -> int:
        return self._block_count

    @property
    def file(self) -> BinaryIO:
        return self._file

    @property
    def path(self) -> str:
        return self._path
